""" REST helpers for Muffin Framework. """

# Package information
# ===================

__version__ = "0.0.2"
__project__ = "muffin-rest"
__author__ = "Kirill Klenov <horneds@gmail.com>"
__license__ = "MIT"

from .forms import *    # noqa
from .handlers import * # noqa



""" Support Muffin-Peewee. """

import muffin
from muffin_rest import RESTHandler


class PWRESTHandler(RESTHandler):

    """ Support REST for Peewee. """

    model = None
    simple = {}

    def get_many(self, request):
        """ Get collection. """
        return self.model.select()

    def get_one(self, request):
        """ Load a resource. """
        resource = request.match_info.get(self.name)
        if not resource:
            return None

        try:
            return self.collection.where(self.model._meta.primary_key == resource).get()
        except Exception:
            raise muffin.HTTPNotFound()

    def populate(self):
        """ Create object. """
        return self.model()

    def save_form(self, form, request, **resources):
        """ Save data. """
        resource = yield from super(PWRESTHandler, self).save_form(form, request, **resources)
        resource.save()
        return resource

    def to_simple(self, data, many=False):
        """ Serialize data. """
        if many:
            return super(PWRESTHandler, self).to_simple(data, many)
        return data.to_simple(**self.simple)

    def delete(self, request, **resources):
        """ Delete a resource. """
        resource = resources.get(self.name)
        if not resource:
            raise muffin.HTTPNotFound('Resource not found')
        resource.delete().execute()



""" REST support. """

import asyncio
import ujson as json
import muffin


class RESTHandlerMeta(type(muffin.Handler)):

    """ Check REST handlers. """

    __coros = 'authorize', 'get_many', 'get_one', 'get_form', 'save_form'

    def __new__(mcs, name, bases, params):
        """ Prepare handler params. """
        cls = super(RESTHandlerMeta, mcs).__new__(mcs, name, bases, params)
        for name in mcs.__coros:
            setattr(cls, name, muffin.to_coroutine(getattr(cls, name)))

        return cls


class RESTHandler(muffin.Handler, metaclass=RESTHandlerMeta):

    """ Implement handler for REST operations. """

    form = None

    @classmethod
    def connect(cls, app, *paths, name=None):
        """ Connect to the application. """
        if not paths:
            paths = [muffin.sre('/%s(/{%s})?/?' % (cls.name, cls.name))]
        return super(RESTHandler, cls).connect(app, *paths, name=name)

    @asyncio.coroutine
    def dispatch(self, request):
        """ Process REST. """
        self.auth = yield from self.authorize(request)
        self.collection = yield from self.get_many(request)
        resources = {}
        if request.method != 'POST':
            resource = yield from self.get_one(request)
            resources[self.name] = resource
        return (yield from super(RESTHandler, self).dispatch(request, **resources))

    def authorize(self, request):
        """ Base point for authorization. """
        return True

    def get_many(self, request):
        """ Base point for collect data. """
        return []

    def get_one(self, request):
        """ Base point load resource. """
        return request.match_info.get(self.name)

    def get_form(self, request, **resources):
        """ Base point load resource. """
        formdata = yield from self.parse(request)
        resource = resources.get(self.name)

        if not self.form:
            raise muffin.MuffinException("%s.form is not defined." % type(self))

        return self.form(formdata, obj=resource)

    def populate(self):
        """ Create object. """
        return object()

    def save_form(self, form, request, **resources):
        """ Save self form. """
        resource = resources.get(self.name, self.populate())
        form.populate_obj(resource)
        return resource

    def to_simple(self, data, many=False):
        """ Convert resource to simple. """
        if many:
            return [self.to_simple(r) for r in data]
        return str(data)

    def get(self, request, **resources):
        """ Get collection of resources. """
        resource = resources.get(self.name)
        if resource:
            return self.to_simple(resource)
        return self.to_simple(self.collection, True)

    def post(self, request):
        """ Create a resource. """
        form = yield from self.get_form(request)
        if not form.validate():
            raise muffin.HTTPBadRequest(
                text=json.dumps(form.errors), content_type='application/json')
        resource = yield from self.save_form(form, request)
        return self.to_simple(resource)

    def put(self, request, **resources):
        """ Update a resource. """
        resource = resources.get(self.name)
        if not resource:
            raise muffin.HTTPNotFound('Resource not found')

        form = yield from self.get_form(request, **resources)
        if not form.validate():
            raise muffin.HTTPBadRequest(
                text=json.dumps(form.errors), content_type='application/json')
        resource = yield from self.save_form(form, request, **resources)
        return self.to_simple(resource)

    patch = put

    def delete(self, request, **resources):
        """ Delete a resource. """
        resource = resources.get(self.name)
        if not resource:
            raise muffin.HTTPNotFound('Resource not found')



from muffin_rest.peewee import PWRESTHandler

from example import app
from example.forms import ResourceForm
from example.models import ResourceModel


@app.register
class Resource(PWRESTHandler):

    form = ResourceForm
    model = ResourceModel



""" Update WTForms for REST. """

from wtforms import validators, widgets         # noqa
from wtforms.fields import *                    # noqa
from wtforms.form import Form as WTForm
from wtforms.validators import ValidationError  # noqa


BooleanField.false_values = BooleanField.false_values + (False,)


class MultiDict(dict):

    """ Muttable multidict. """

    def getlist(self, key):
        """ Implement method. """
        val = self[key]
        if not isinstance(val, list):
            val = [val]
        return val

    def getall(self, key):
        """ Implement method. """
        return [self[key]]


class Form(WTForm):

    """ Supports partitial updates. """

    def __init__(self, formdata=None, obj=None, prefix='', data=None, meta=None, **kwargs):
        """ Convert formdata to multidict. """
        if formdata is not None:
            formdata = MultiDict(formdata)

        super(Form, self).__init__(formdata, obj, prefix, data, meta, **kwargs)

    def process(self, formdata=None, obj=None, data=None, **kwargs):
        """ Fill formdata from object. And support `process_field` method. """
        for name in self._fields:
            if name in formdata:
                if hasattr(self, 'process_%s' % name):
                    formdata[name] = getattr(self, 'process_%s' % name)(formdata[name])
            elif obj:
                formdata[name] = getattr(obj, name, None)

        super(Form, self).process(formdata, obj, data, **kwargs)
